# :coding: utf-8

import wiz.command_line


def main():
    """Execute main command line interface passing command line arguments."""
    wiz.command_line.main(prog_name="wiz")


if __name__ == "__main__":
    main()
